const jmespath = require("jmespath");

/**

 * @param {Object} project Project object.
 */
const createProject = (knex, projectName) =>
  knex.transaction(async (trx) => {
    try {
      const projectRes = await trx("projects")
      .where("name", "=", projectName)
      .first();
      
      if (projectRes) {
        console.log(`${projectName} already exists`);
      }
      
      return trx("projects")
      .insert({
        name: projectName,
      })
      .onConflict("name")
      .ignore();
    } catch (err) {
      console.log(err);
    }
  });

/**
 * @param knex
 * @param {Object} run Run object.
 */
const createRun = (knex, run, projectName) =>
  knex.transaction(async (trx) => {
    try {
      const project = await trx("projects")
      .where("name", "=", projectName)
      .first();
      
      return trx("runs").insert({
        runName: run.runName,
        suites: run.suites,
        tests: run.tests,
        passes: run.passes,
        failures: run.failures,
        pending: run.pending,
        start: run.start,
        end: run.end,
        duration: run.duration,
        testsRegistered: run.testsRegistered,
        passPercent: run.passPercent,
        pendingPercent: run.pendingPercent,
        other: run.other,
        hasOther: run.hasOther,
        skipped: run.skipped,
        hasSkipped: run.hasSkipped,
        projectId: project.id
      });
    } catch (err) {
      console.log(err);
    }
  });

const deleteProject = (knex) =>
  knex.transaction(async (trx) => {
    try {
      return trx("projects").del()
    } catch (err) {
      console.log(err);
    }
  });

const createResult = (knex, runId, result) =>
  knex.transaction(async (trx) => {
    try {
      return trx("results").insert({
        runId: runId,
        title: result.title,
        file: result.file,
        fullFile: result.fullFile,
        jsonUuid: result.uuid
      });
    } catch (err) {
      console.log(err);
    }
  });

const createSuite = (knex, suite) =>
  knex.transaction(async (trx) => {
    try {
      return trx("suites").insert({
        resultId: suite.resultId,
        fullFile: suite.fullFile,
        duration: suite.duration,
        jsonUuid: suite.jsonUuid
      });
    } catch (err) {
      console.log(err);
    }
  });

const createTest = (knex, test) =>
  knex.transaction(async (trx) => {
    try {
      return trx("tests").insert({
        suiteId: test.suiteId,
        title: test.title,
        fullTitle: test.fullTitle,
        duration: test.duration,
        state: test.state,
        pass: test.pass,
        fail: test.fail,
        pending: test.pending,
        context: test.context,
        isHook: test.isHook,
        skipped: test.skipped,
        jsonUuid: test.jsonUuid
      });
    } catch (err) {
      console.log(err);
    }
  });

const createError = (knex, error) =>
  knex.transaction(async (trx) => {
    try {
      return trx("errors").insert({
        testId: error.testId,
        message: error.message,
        estack: error.estack,
      });
    } catch (err) {
      console.log(err);
    }
  });

const getRun = (knex, run) =>
  knex.transaction(async (trx) => {
    try {
      return (run = await trx("runs")
      .select("id")
      .orderBy("createdAt", "desc")
      .first());
    } catch (err) {
      console.log(err);
    }
  });

const getUuidResult = (knex, parentJsonUuid, resultUuid) =>
  knex.transaction(async (trx) => {
    try {
      return (resultUuid = await trx("results")
        .select("id")
        .where("jsonUuid", "=", parentJsonUuid)
        .first()
      )
    } catch (err) {
      console.log(err);
    }
  });

const getUuidSuite = (knex, parentJsonUuid, suiteUuid) =>
  knex.transaction(async (trx) => {
    try {
      return (suiteUuid = await trx("suites")
        .select("id")
        .where("jsonUuid", "=", parentJsonUuid)
        .first()
      )
    } catch (err) {
      console.log(err);
    }
  });

const getUuidTest = (knex, parentJsonUuid, testUuid) =>
  knex.transaction(async (trx) => {
    try {
      return (testUuid = await trx("tests")
        .select("id")
        .where("jsonUuid", "=", parentJsonUuid)
        .first()
      )
    } catch (err) {
      console.log(err);
    }
  });


async function insertResults(knex, run, results){
  for (const result of results) {
    await createResult(knex, run.id, result).then(() => {
      result.suites.forEach((suite) => {
        suite['parentJsonUuid'] = result.uuid
      })
    })
  }
  return results
}

async function insertSuites(knex,suites){
  for (const suite of suites) {
    await createSuite(knex,{
      resultId: suite.resultId,
      fullFile: suite.fullFile,
      duration: suite.duration,
      jsonUuid: suite.uuid
    }).then(() => {
      suite.tests.forEach((test) => {
        test['parentJsonUuid'] = suite.uuid
      })
      
    })
  }
  return suites
}

async function insertTests(knex, tests){
  for (const test of tests) {
    await createTest(knex,{
      suiteId: test.suiteId,
      title: test.title,
      fullTitle: test.fullTitle,
      duration: test.duration,
      state: test.state,
      pass: test.pass,
      fail: test.fail,
      pending: test.pending,
      context: test.context,
      isHook: test.isHook,
      skipped: test.skipped,
      jsonUuid: test.uuid
    }).then(() => {
      test.err['parentJsonUuid'] = test.uuid
    })
  }
  return tests
}

async function insertErrors(knex, errors){
  for (const error of errors) {
    await createError(knex,{
      testId: error.testId,
      message: error.message,
      estack: error.estack,
    })
  }
  return errors
}

async function addResultIdSuites(knex, results, suites){
  suites = jmespath.search(
    results,
    `[].suites`
  ).flat()
  for (const suite of suites) {
    //console.log('Parent Result: ' + suite.parentJsonUuid)
    const res = await getUuidResult(knex, suite.parentJsonUuid)
    suite['resultId'] = res.id
  }
  return suites
}

async function addSuiteIdTests(knex, suites, tests){
  tests = jmespath.search(
    suites,
    `[].tests`
  ).flat()
  for (const test of tests) {
    //console.log('Parent Suite: ' + test.parentJsonUuid)
    const suite = await getUuidSuite(knex, test.parentJsonUuid)
    test['suiteId'] = suite.id
  }
  return tests
}

async function addTestIdErrors(knex, tests, errors){
  errors = jmespath.search(
    tests,
    `[].err`
  ).flat()
  for (const error of errors) {
    //console.log('Parent Test: ' + error.parentJsonUuid)
    const test = await getUuidTest(knex, error.parentJsonUuid)
    error['testId'] = test.id
  }
  return errors
}


async function inserter(knex, projectName, jsonReport){
  
  let run = jsonReport.stats;
  run["runName"] = "Run at:" + run.start;
  run["projectName"] = projectName;
  
  let results = jmespath.search(
    jsonReport,
    `results[].{ title:title, file:file, fullFile:fullFile, suites:suites , uuid:uuid}`
  )
  
  await deleteProject(knex);
  await createProject(knex, projectName)
  await createRun(knex, run, projectName)
  const r = await getRun(knex, run)
  results = await insertResults(knex, r, results)
  let suites = await addResultIdSuites(knex, results)
  suites = await insertSuites(knex, suites)
  let tests = await addSuiteIdTests(knex, suites)
  tests = await insertTests(knex, tests)
  let errors = await addTestIdErrors(knex, tests)
  await insertErrors(knex, errors)
  
  await knex.destroy()
  console.log("The report has been inserted")
  console.log("Disconnect to DB")
}



module.exports = {
  inserter
};
